use core::fmt::{Display, Formatter};

use std::collections::HashMap;
use std::error::Error as ErrorTrait;
use std::fs::File;
use std::path::PathBuf;

use crate::classfile::JavaClassFile;
use crate::classfile;

#[derive(Debug)]
pub struct ClassStore {
    class_ids: HashMap<String, usize>,
    classes: Vec<(bool, JavaClassFile)>, // was_init, class_file
    class_path_fragments: Vec<PathBuf>,
}

#[derive(Debug)]
pub enum Error {
    ClassNotFoundError(String),
    IOError(std::io::Error),
    ClassFileError(String, classfile::Error),
}

impl From<std::io::Error> for Error {
    fn from(value: std::io::Error) -> Self {
        return Error::IOError(
            value
        );
    }
}

impl From<classfile::Error> for Error {
    fn from(value: classfile::Error) -> Self {
        return Error::ClassFileError(
            "An error occured while loading a classfile".to_string(),
            value
        );
    }
}

impl ErrorTrait for Error {}
impl Display for Error {
    fn fmt(&self, formatter: &mut Formatter<'_>) -> Result<(), std::fmt::Error> {
            writeln!(formatter, "{self}")?;
            if let Some(e) = self.source() {
                writeln!(formatter, "\tCaused by: {e:?}")?;
            }
            Ok(())
    }
}

impl ClassStore {
    pub fn new() -> Self {
        let current_dir_path = PathBuf::from("./");

        ClassStore {
            class_ids: HashMap::new(),
            classes: Vec::new(),
            class_path_fragments: vec![current_dir_path],
        }
    }

    pub fn add_class(&mut self, class_file: JavaClassFile, was_init: bool) {
        return self.classes.push((was_init, class_file));
    }

    pub fn load_class_from_file(&mut self, class_file_path: &PathBuf) -> Result<usize, Error> {
        let mut file_reader = File::open(class_file_path)?;
        let classfile = JavaClassFile::new(&mut file_reader)?;

        let classname = classfile.get_classname()?;
        self.class_ids.insert(classname, self.classes.len());
        self.classes.push((false, classfile));

        return Ok(self.classes.len() - 1);
    }

    pub fn load_class(&mut self, classname: &String) -> Result<usize, Error> {
        let mut path_buf = PathBuf::new();

        for class_path in &self.class_path_fragments {
            path_buf.push(class_path);
            path_buf.push(&classname);
            path_buf.set_extension("class");

            if path_buf.is_file() {
                return self.load_class_from_file(&path_buf);
            }
        };

        return Err(Error::ClassNotFoundError(format!("Could not find class '{}' in classpath", classname)));
    }

    pub fn have_class(&self, classname: &String) -> bool {
        return self.class_ids.contains_key(classname);
    }

    pub fn get_class(&self, classname: &String) -> Result<(&JavaClassFile, usize), Error> {
        let class_id = self.class_ids.get(classname);

        return match class_id {
            Some(id) => Ok((&self.classes[*id].1, *id)),
            None => Err(Error::ClassNotFoundError(format!("Could not locate class '{}'", classname))),
        }
    }

    pub fn get_or_load_class(&mut self, classname: &String) -> Result<(&JavaClassFile, usize), Error> {
        if self.have_class(classname) {
            return Ok(self.get_class(classname)?);
        } else {
            let class_idx = self.load_class(classname)?;
            return Ok((&self.classes[class_idx].1, class_idx));
        }
    }

    pub fn class_file_from_idx(&self, idx: usize) -> Option<&JavaClassFile> {
        return match self.classes.get(idx) {
            Some((_was_init, class_file)) => Some(class_file),
            None => None,
        }
    }

    pub fn class_idx_from_name(&self, classname: &String) -> Option<&usize> {
        return self.class_ids.get(classname);
    }

    pub fn was_init(&self, classname: &String) -> Option<&bool> {
        let (was_init, _) = self.classes.get(*self.class_idx_from_name(classname).unwrap()).unwrap();

        return Some(was_init);
    }
}


